package java.io;

import java.util.Locale;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;

class WinNTFileSystem extends FileSystem
{
    private final char slash;
    private final char altSlash;
    private final char semicolon;
    private static String[] driveDirCache;
    private ExpiringCache cache;
    private ExpiringCache prefixCache;
    
    public WinNTFileSystem() {
        this.cache = new ExpiringCache();
        this.prefixCache = new ExpiringCache();
        this.slash = ((String)AccessController.doPrivileged(new GetPropertyAction("file.separator"))).charAt(0);
        this.semicolon = ((String)AccessController.doPrivileged(new GetPropertyAction("path.separator"))).charAt(0);
        this.altSlash = ((this.slash == '\\') ? '/' : '\\');
    }
    
    private boolean isSlash(char c) {
        return c == '\\' || c == '/';
    }
    
    private boolean isLetter(char c) {
        return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z');
    }
    
    private String slashify(String str) {
        if (str.length() > 0 && str.charAt(0) != this.slash) {
            return this.slash + str;
        }
        return str;
    }
    
    @Override
    public char getSeparator() {
        return this.slash;
    }
    
    @Override
    public char getPathSeparator() {
        return this.semicolon;
    }
    
    @Override
    public String normalize(String s) {
        int length = s.length();
        int slash = this.slash;
        int altSlash = this.altSlash;
        int n = '\0';
        int i = 0;
        while (i < length) {
            int char1 = s.charAt(i);
            if (char1 == altSlash) {
                return this.normalize(s, length, (n == slash) ? (i - 1) : i);
            }
            if (char1 == slash && n == slash && i > 1) {
                return this.normalize(s, length, i - 1);
            }
            if (char1 == ':' && i > 1) {
                return this.normalize(s, length, 0);
            }
            n = char1;
            ++i;
        }
        if (n == slash) {
            return this.normalize(s, length, length - 1);
        }
        return s;
    }
    
    private String normalize(String s, int capacity, int endIndex) {
        if (capacity == 0) {
            return s;
        }
        if (endIndex < 3) {
            endIndex = 0;
        }
        int slash = this.slash;
        StringBuffer o = new StringBuffer(capacity);
        int i;
        if (endIndex == 0) {
            i = this.normalizePrefix(s, capacity, o);
        }
        else {
            i = endIndex;
            o.append(s.substring(0, endIndex));
        }
        while (i < capacity) {
            int char1 = s.charAt(i++);
            if (this.isSlash((char)char1)) {
                while (i < capacity && this.isSlash(s.charAt(i))) {
                    ++i;
                }
                if (i == capacity) {
                    int length = o.length();
                    if (length == 2 && o.charAt(1) == ':') {
                        o.append((char)slash);
                        break;
                    }
                    if (length == 0) {
                        o.append((char)slash);
                        break;
                    }
                    if (length == 1 && this.isSlash(o.charAt(0))) {
                        o.append((char)slash);
                        break;
                    }
                    break;
                }
                else {
                    o.append((char)slash);
                }
            }
            else {
                o.append((char)char1);
            }
        }
        return o.toString();
    }
    
    private int normalizePrefix(String s, int n, StringBuffer sb) {
        int n2;
        for (n2 = 0; n2 < n && this.isSlash(s.charAt(n2)); ++n2) {}
        int char1;
        if (n - n2 >= 2 && this.isLetter((char)(char1 = s.charAt(n2))) && s.charAt(n2 + 1) == ':') {
            sb.append((char)char1);
            sb.append(':');
            n2 += 2;
        }
        else {
            n2 = 0;
            if (n >= 2 && this.isSlash(s.charAt(0)) && this.isSlash(s.charAt(1))) {
                n2 = 1;
                sb.append(this.slash);
            }
        }
        return n2;
    }
    
    @Override
    public int prefixLength(String s) {
        int slash = this.slash;
        int length = s.length();
        if (length == 0) {
            return 0;
        }
        int char1 = s.charAt(0);
        int n = (length > 1) ? s.charAt(1) : '\0';
        if (char1 == slash) {
            if (n == slash) {
                return 2;
            }
            return 1;
        }
        else if (this.isLetter((char)char1) && n == ':') {
            if (length > 2 && s.charAt(2) == slash) {
                return 3;
            }
            return 2;
        }
        else {
            return 0;
        }
    }
    
    @Override
    public String resolve(String s, String s2) {
        int length = s.length();
        if (length == 0) {
            return s2;
        }
        int length2 = s2.length();
        if (length2 == 0) {
            return s;
        }
        int srcBegin = 0;
        if (length2 > 1 && s2.charAt(0) == this.slash) {
            if (s2.charAt(1) == this.slash) {
                srcBegin = 2;
            }
            else {
                srcBegin = 1;
            }
            if (length2 == srcBegin) {
                if (s.charAt(length - 1) == this.slash) {
                    return s.substring(0, length - 1);
                }
                return s;
            }
        }
        if (s.charAt(length - 1) == this.slash) {
            --length;
        }
        int n = length + length2 - srcBegin;
        char[] value;
        if (s2.charAt(srcBegin) == this.slash) {
            value = new char[n];
            s.getChars(0, length, value, 0);
            s2.getChars(srcBegin, length2, value, length);
        }
        else {
            value = new char[n + 1];
            s.getChars(0, length, value, 0);
            value[length] = this.slash;
            s2.getChars(srcBegin, length2, value, length + 1);
        }
        return new String(value);
    }
    
    @Override
    public String getDefaultParent() {
        return "" + this.slash;
    }
    
    @Override
    public String fromURIPath(String s) {
        if (s.length() > 2 && s.charAt(2) == ':') {
            s = s.substring(1);
            if (s.length() > 3 && s.endsWith("/")) {
                s = s.substring(0, s.length() - 1);
            }
        }
        else if (s.length() > 1 && s.endsWith("/")) {
            s = s.substring(0, s.length() - 1);
        }
        return s;
    }
    
    @Override
    public boolean isAbsolute(File file) {
        int prefixLength = file.getPrefixLength();
        return (prefixLength == 2 && file.getPath().charAt(0) == this.slash) || prefixLength == 3;
    }
    
    @Override
    public String resolve(File file) {
        String path = file.getPath();
        int prefixLength = file.getPrefixLength();
        if (prefixLength == 2 && path.charAt(0) == this.slash) {
            return path;
        }
        if (prefixLength == 3) {
            return path;
        }
        if (prefixLength == 0) {
            return this.getUserPath() + this.slashify(path);
        }
        if (prefixLength == 1) {
            String userPath2 = this.getUserPath();
            String drive2 = this.getDrive(userPath2);
            if (drive2 != null) {
                return drive2 + path;
            }
            return userPath2 + path;
        }
        else if (prefixLength == 2) {
            String userPath = this.getUserPath();
            String drive = this.getDrive(userPath);
            if (drive != null && path.startsWith(drive)) {
                return userPath + this.slashify(path.substring(2));
            }
            int char1 = path.charAt(0);
            String driveDirectory = this.getDriveDirectory((char)char1);
            if (driveDirectory != null) {
                String string = new StringBuilder().append((char)char1).append(':').append(driveDirectory).append(this.slashify(path.substring(2))).toString();
                SecurityManager securityManager = System.getSecurityManager();
                try {
                    if (securityManager != null) {
                        securityManager.checkRead(string);
                    }
                }
                catch (SecurityException ex) {
                    throw new SecurityException("Cannot resolve path " + path);
                }
                return string;
            }
            return (char)char1 + ":" + this.slashify(path.substring(2));
        }
        else {
            throw new InternalError("Unresolvable path: " + path);
        }
    }
    
    private String getUserPath() {
        return this.normalize(System.getProperty("user.dir"));
    }
    
    private String getDrive(String s) {
        return (this.prefixLength(s) == 3) ? s.substring(0, 2) : null;
    }
    
    private static int driveIndex(char c) {
        if (c >= 'a' && c <= 'z') {
            return c - 97;
        }
        if (c >= 'A' && c <= 'Z') {
            return c - 65;
        }
        return -1;
    }
    
    private native String getDriveDirectory(int p0);
    
    private String getDriveDirectory(char c) {
        int driveIndex = driveIndex(c);
        if (driveIndex < 0) {
            return null;
        }
        String s = WinNTFileSystem.driveDirCache[driveIndex];
        if (s != null) {
            return s;
        }
        return WinNTFileSystem.driveDirCache[driveIndex] = this.getDriveDirectory(driveIndex + 1);
    }
    
    @Override
    public String canonicalize(String s) throws IOException {
        int length = s.length();
        if (length == 2 && this.isLetter(s.charAt(0)) && s.charAt(1) == ':') {
            int char2 = s.charAt(0);
            if (char2 >= 'A' && char2 <= 'Z') {
                return s;
            }
            return "" + (char)(char2 - 32) + 58;
        }
        else if (length == 3 && this.isLetter(s.charAt(0)) && s.charAt(1) == ':' && s.charAt(2) == '\\') {
            int char1 = s.charAt(0);
            if (char1 >= 'A' && char1 <= 'Z') {
                return s;
            }
            return "" + (char)(char1 - 32) + 58 + 92;
        }
        else {
            if (!WinNTFileSystem.useCanonCaches) {
                return this.canonicalize0(s);
            }
            String s2 = this.cache.get(s);
            if (s2 == null) {
                String parentOrNull = null;
                if (WinNTFileSystem.useCanonPrefixCache) {
                    parentOrNull = parentOrNull(s);
                    if (parentOrNull != null) {
                        String value = this.prefixCache.get(parentOrNull);
                        if (value != null) {
                            String substring = s.substring(1 + parentOrNull.length());
                            s2 = this.canonicalizeWithPrefix(value, substring);
                            this.cache.put(parentOrNull + File.separatorChar + substring, s2);
                        }
                    }
                }
                if (s2 == null) {
                    s2 = this.canonicalize0(s);
                    this.cache.put(s, s2);
                    if (WinNTFileSystem.useCanonPrefixCache && parentOrNull != null) {
                        String parentOrNull2 = parentOrNull(s2);
                        if (parentOrNull2 != null) {
                            File o = new File(s2);
                            if (o.exists() && !o.isDirectory()) {
                                this.prefixCache.put(parentOrNull, parentOrNull2);
                            }
                        }
                    }
                }
            }
            return s2;
        }
    }
    
    private native String canonicalize0(String p0) throws IOException;
    
    private String canonicalizeWithPrefix(String str, String str2) throws IOException {
        return this.canonicalizeWithPrefix0(str, str + File.separatorChar + str2);
    }
    
    private native String canonicalizeWithPrefix0(String p0, String p1) throws IOException;
    
    private static String parentOrNull(String s) {
        if (s == null) {
            return null;
        }
        int separatorChar = File.separatorChar;
        int n = '/';
        int i = s.length() - 1;
        int n2 = 0;
        int n3 = 0;
        while (i > 0) {
            int char1 = s.charAt(i);
            if (char1 == '.') {
                if (++n2 >= 2) {
                    return null;
                }
                if (n3 == 0) {
                    return null;
                }
            }
            else if (char1 == separatorChar) {
                if (n2 == 1 && n3 == 0) {
                    return null;
                }
                if (i == 0 || i >= i - 1 || s.charAt(i - 1) == separatorChar || s.charAt(i - 1) == n) {
                    return null;
                }
                return s.substring(0, i);
            }
            else {
                if (char1 == n) {
                    return null;
                }
                if (char1 == '*' || char1 == '?') {
                    return null;
                }
                ++n3;
                n2 = 0;
            }
            --i;
        }
        return null;
    }
    
    @Override
    public native int getBooleanAttributes(File p0);
    
    @Override
    public native boolean checkAccess(File p0, int p1);
    
    @Override
    public native long getLastModifiedTime(File p0);
    
    @Override
    public native long getLength(File p0);
    
    @Override
    public native boolean setPermission(File p0, int p1, boolean p2, boolean p3);
    
    @Override
    public native boolean createFileExclusively(String p0) throws IOException;
    
    @Override
    public native String[] list(File p0);
    
    @Override
    public native boolean createDirectory(File p0);
    
    @Override
    public native boolean setLastModifiedTime(File p0, long p1);
    
    @Override
    public native boolean setReadOnly(File p0);
    
    @Override
    public boolean delete(File file) {
        this.cache.clear();
        this.prefixCache.clear();
        return this.delete0(file);
    }
    
    private native boolean delete0(File p0);
    
    @Override
    public boolean rename(File file, File file2) {
        this.cache.clear();
        this.prefixCache.clear();
        return this.rename0(file, file2);
    }
    
    private native boolean rename0(File p0, File p1);
    
    @Override
    public File[] listRoots() {
        int listRoots0 = listRoots0();
        int n = 0;
        for (int i = 0; i < 26; ++i) {
            if ((listRoots0 >> i & 0x1) != 0x0) {
                if (!this.access((char)(65 + i) + ":" + this.slash)) {
                    listRoots0 &= ~(1 << i);
                }
                else {
                    ++n;
                }
            }
        }
        File[] array = new File[n];
        int n2 = 0;
        int slash = this.slash;
        for (int j = 0; j < 26; ++j) {
            if ((listRoots0 >> j & 0x1) != 0x0) {
                array[n2++] = new File((char)(65 + j) + ":" + (char)slash);
            }
        }
        return array;
    }
    
    private static native int listRoots0();
    
    private boolean access(String file) {
        try {
            SecurityManager securityManager = System.getSecurityManager();
            if (securityManager != null) {
                securityManager.checkRead(file);
            }
            return true;
        }
        catch (SecurityException ex) {
            return false;
        }
    }
    
    @Override
    public long getSpace(File file, int n) {
        if (file.exists()) {
            return this.getSpace0(file, n);
        }
        return 0L;
    }
    
    private native long getSpace0(File p0, int p1);
    
    @Override
    public int compare(File file, File file2) {
        return file.getPath().compareToIgnoreCase(file2.getPath());
    }
    
    @Override
    public int hashCode(File file) {
        return file.getPath().toLowerCase(Locale.ENGLISH).hashCode() ^ 0x12D591;
    }
    
    private static native void initIDs();
    
    static {
        WinNTFileSystem.driveDirCache = new String[26];
        initIDs();
    }
}
